#include "nngfs.h"


static inline Indirect *nngfs_get_branch(struct inode *inode,int depth,int *offsets,
                                         Indirect chain[2],int *err){
 // struct super_block *sb = inode->i_sb;
  Indirect *p = chain;
  //struct buffer_head *bh;
  *err = 0;
  NNGFS_I(inode)->blockno = inode->i_ino;
  add_chain(chain,NULL,&(NNGFS_I(inode)->blockno));
  if(!p->key)
   goto no_block;
  while(--depth){
     printk(KERN_INFO "NGFS : depth = %d in %s,something wrong\n",depth,__FUNCTION__);
    //Nothing to write as of now as depth is always one.....
  }
  return NULL;
no_block:
   return p;
}

int nngfs_block_to_path(struct inode *inode,sector_t block,int offsets[2],int *err){
   int n=0;
   /* block should always be 0 for NNGFS as it is a single block filesystem */
   if(block < 0 )
     printk(KERN_INFO "NGFS: block [%d] less than 0 in %s \n",(int)block,__FUNCTION__);
   else if(block>0) {
     printk(KERN_INFO "NGFS: block [%d] greater than 0 in %s \n",(int)block,__FUNCTION__);
   } else {
     offsets[n++] = block;
   }
   return n;
}

int nngfs_new_block(struct inode* inode){
  struct nngfs_superblock *n_sb;
  struct super_block *sb = inode->i_sb;
  struct nngfs_info *nsi = (struct nngfs_info *)sb->s_fs_info;
  struct buffer_head *bh;
   n_sb = (struct nngfs_superblock *)nsi->ngfs_sb;
   bh = (struct buffer_head *)(nsi->sbh);
   n_sb->block_bitmap = setbit(n_sb->block_bitmap,inode->i_ino+1);
   mark_buffer_dirty(bh);
   return inode->i_ino+1;
}

int nngfs_alloc_branch(struct inode *inode,int num,int *offsets,Indirect *branch){
  int parent = nngfs_new_block(inode);
  branch[0].key = parent;
  return 0;
}

int nngfs_splice_branch(struct inode *inode,Indirect chain[2], Indirect *where,int num){
   mark_inode_dirty(inode);
   return 0;
}


int nngfs_get_block(struct inode *inode,sector_t block, struct buffer_head *bh_result,int create){
 int err =-EIO;
 int offsets[2];
 Indirect chain[2];
 Indirect *partial;
 int left,goal;
 int depth = nngfs_block_to_path(inode,block,offsets,&err);
 goal = 0;
 memset(chain,0,sizeof(Indirect)*2);
 if(depth == 0)
  goto out;
 reread:
  partial = nngfs_get_branch(inode,depth,offsets,chain,&err);
  if(!partial){
got_it:
   map_bh(bh_result,inode->i_sb,chain[depth-1].key);
   partial = chain+depth-1;
   goto cleanup;
}
 if(!create || err == -EIO ){
cleanup:
   while(partial > chain){
      brelse(partial->bh);
      partial--;
    }
out:
   return err;
 }

 if(err == -EAGAIN)
   goto changed;

 left = (chain+depth) - partial;
 err = nngfs_alloc_branch(inode,left,offsets+(partial-chain),partial);
 if(err)
   goto cleanup;

  if(nngfs_splice_branch(inode,chain,partial,left) < 0)
     goto changed;

  set_buffer_new(bh_result);
  goto got_it;

changed:
   while(partial > chain) {
    brelse(partial->bh);
    partial--;
   }
   goto reread;
}


int nngfs_prepare_write(struct file *file,struct page *page,unsigned from,unsigned to){
      struct super_block *sb = file->f_dentry->d_inode->i_sb;
      if(!from){
         if(to >= sb->s_blocksize)
           return -ENOSPC;
      }
      else if (file->f_dentry->d_inode->i_size + (to - from) >= sb->s_blocksize )
                return -ENOSPC;

      return block_prepare_write(page,from,to,nngfs_get_block);
}

int nngfs_readpage(struct file *file,struct page *page){
      return mpage_readpage(page,nngfs_get_block);
}

int nngfs_writepage(struct page *page,struct writeback_control *wbc){
      return block_write_full_page(page,nngfs_get_block,wbc);
}

sector_t nngfs_bmap(struct address_space *mapping,sector_t block){
      return generic_block_bmap(mapping,block,nngfs_get_block);
}

struct address_space_operations nngfs_aops = {
        readpage:         nngfs_readpage,
        writepage:        nngfs_writepage,
 //       sync_page:        block_sync_page,
        prepare_write:    nngfs_prepare_write,
        commit_write:     generic_commit_write,
        bmap:             nngfs_bmap,
};
