#include "nngfs.h"


extern unsigned32 clear_dentry(struct inode *nip ,int inum);
extern struct inode_operations nngfs_dir_inode_operations;
extern struct file_operations nngfs_dir_operations;
extern struct address_space_operations nngfs_aops;
extern struct file_operations nngfs_file_operations;
struct inode_operations nngfs_file_inode_operations;
extern int nngfs_dir_add(struct inode *dir,unsigned32 inum,const unsigned char *name);

unsigned32 Get_free_inode_number (struct nngfs_superblock * sb){
        unsigned32 i,bitmap = sb->inode_bitmap;
        for(i=0;i<sizeof(bitmap)*8;i++)
              if(!(bitmap & (1<<i))){
                  return i+1;
              }
        return -1;
}

unsigned32 nngfs_find_entry(struct inode *nip,const unsigned char *name){
       unsigned32 inum = nip->i_ino; //this is the inode of directory. The corresponding block is also at the same place.
       struct super_block *sb = nip->i_sb;
       struct buffer_head *bh ;
       struct nngfs_dir_entry *de;;
       bh = sb_bread(sb,inum+1);
       if(!bh){
          printk(KERN_INFO "NGFS (%s) : Could not read block [%d] \n",__FUNCTION__,inum+1);
          return -1;
       }
       de = (struct nngfs_dir_entry *)bh->b_data;
       while(strcmp(de->name,"")){
           if(!strcmp(de->name,name) && de->inode != NNGFS_INUM_INVALID){
             brelse(bh);
             return de->inode;
           }
         de = (struct nngfs_dir_entry *)((char *)de + de->rec_len);
       }
       brelse(bh);
       return -1;
}

int nngfs_unlink(struct inode * dir, struct dentry *dentry){
  int err = -ENOENT,ret;
  struct buffer_head *bh;
  struct nngfs_superblock *n_sb;
  struct super_block *sb = dir->i_sb;
  struct nngfs_info *nsi = (struct nngfs_info *)sb->s_fs_info;
  n_sb = (struct nngfs_superblock *)nsi->ngfs_sb;
  bh = (struct buffer_head *)(nsi->sbh);
  //printk(KERN_INFO "NNGFS (%s) : Called \n",__FUNCTION__);

   //See if the entry exists in the dir
       ret = nngfs_find_entry(dir,dentry->d_name.name);
       if(ret!=-1){
         clear_dentry(dir,ret);
        // printk(KERN_INFO "Freeing inode %d\n",ret);
         //printk(KERN_INFO "Current bitmap 0x%x \n",n_sb->inode_bitmap);
         n_sb->inode_bitmap = clearbit(n_sb->inode_bitmap,ret);
         //printk(KERN_INFO "Changed bitmap 0x%x \n",n_sb->inode_bitmap);
         mark_buffer_dirty(bh);
         inode_dec_link_count(dentry->d_inode);
         return 0;
       }
  return err;
}

int nngfs_permission(struct inode *inode,int mask,struct nameidata *nd){
       return 0;
}

struct dentry *nngfs_lookup(struct inode *inode,struct dentry *de,struct nameidata *nd){
       // struct nngfs_inode *nip = (struct nngfs_inode *)(inode->i_private);
        int inum;
        struct inode *newino = NULL;
        if(de->d_name.len > NNGFS_NAME_LEN){
           return ERR_PTR(-ENAMETOOLONG);
        }
        de->d_op = inode->i_sb->s_root->d_op;
        inum = nngfs_find_entry(inode,de->d_name.name);
        if(inum != -1){
         newino = iget(inode->i_sb,inum);
         if(!newino)
            return ERR_PTR(-EACCES);
        }
         d_add(de,newino);
         return NULL;
}

int nngfs_create(struct inode *dir, struct dentry *dentry, int mode, struct nameidata *nd)
{
       int inum=0,ret=0;
       struct buffer_head *bh;
       struct inode *newnip;
       struct nngfs_inode_info *nino;
       struct super_block *sb = dir->i_sb;
       struct nngfs_info *nsb = (struct nngfs_info *)(sb->s_fs_info);
       struct nngfs_superblock *n_sb = nsb->ngfs_sb;

       //See if the entry already exists in the dir
       ret = nngfs_find_entry(dir,dentry->d_name.name);
       if(ret!=-1){
          return PTR_ERR(ERR_PTR(-EEXIST));
       }

       //See the bitmap and get the inode number
       inum = Get_free_inode_number(n_sb);
       if(inum == -1){
        printk(KERN_INFO "NGFS (%s) :Got free inode number = %d\n",__FUNCTION__,inum);
        return PTR_ERR(ERR_PTR(-ENOSPC));
       }

       //Allocate the new inode
       newnip = new_inode(sb);
       if(!newnip){
        printk(KERN_INFO "NGFS (%s) :Could not allocate free inode\n",__FUNCTION__);
        return PTR_ERR(ERR_PTR(-ENOSPC));
       }


       bh = (struct buffer_head *)(nsb->sbh);
       n_sb->inode_bitmap=setbit(n_sb->inode_bitmap,inum);
       newnip->i_ino = inum;
       dir->i_nlink++;

       newnip->i_atime = newnip->i_mtime = newnip->i_ctime = CURRENT_TIME_SEC;
       newnip->i_uid = dir->i_uid;
       newnip->i_gid = dir->i_gid;
      if(S_ISDIR(mode)){
       newnip->i_mode =  NNGFS_DIR;
      newnip->i_op = &nngfs_dir_inode_operations;
       newnip->i_fop = &nngfs_dir_operations;
     }
     else{
       newnip->i_mode =  NNGFS_FILE;
       newnip->i_op = &nngfs_file_inode_operations;
       newnip->i_fop = &nngfs_file_operations;
     }

      newnip->i_mapping->a_ops = &nngfs_aops;

       nino = NNGFS_I(newnip);
       
       nino->inode = inum;
       nino->filetype = newnip->i_mode ;
       nino->atime = nino->mtime = nino->ctime = CURRENT_TIME_SEC;
       nino->uid = dir->i_uid;
       nino->gid = dir->i_gid;
       nino->size = 0; 
       nino->blockno = nino->inode+1;
       nino->magic = 0xdeadbeef;
       
       sb->s_dirt=1;
      // inode_inc_link_count(newnip);
       insert_inode_hash(newnip);
       d_instantiate(dentry,newnip);
       mark_buffer_dirty(bh); //Write the new bitmap to disk
       mark_inode_dirty(newnip);
       //Add the name information to dir entry
       if(nngfs_dir_add(dir,inum,dentry->d_name.name)){
        printk(KERN_INFO "NGFS (%s) :Could not allocate free inode\n",__FUNCTION__);
        return PTR_ERR(ERR_PTR(-EMFILE));
       }
       mark_inode_dirty(dir);
       return 0;
}

struct inode_operations nngfs_file_inode_operations ={
     .permission = nngfs_permission,
};
