/* 
 * nngfs/nngfs.h
 * 
 * Copyright (C) 2008
 * Manish Katiyar (mkatiyar@gmail.com) 
 *
 */

#ifndef __NNGFS__
#define __NNGFS__
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/buffer_head.h>
#include <linux/blkdev.h>
#include <linux/slab.h>
#include <linux/mpage.h>
#include <linux/init.h>
#include <linux/parser.h>

#define NNGFS_MAGIC 0x4E474653  /* "NGFS" */
#define NNGFS_NAME_LEN 16
#define NNGFS_SUPERBLOCK 0
#define NNGFS_BLOCK_SIZE 4096
#define NNGFS_INODE_BLOCK 1
#define NNGFS_INUM_INVALID -5
#define NNGFS_CLEAN 0
#define NNGFS_DIRTY ~(NNGFS_CLEAN)
#define NNGFS_ROOT_INODE 1
#define NNGFS_FILE (S_IFREG|S_IRUSR|S_IWUSR|S_IXUSR|S_IROTH|S_IXOTH|S_IRGRP|S_IXGRP)
#define NNGFS_DIR  (S_IFDIR|S_IRUSR|S_IWUSR|S_IXUSR|S_IROTH|S_IXOTH|S_IRGRP|S_IXGRP)
#define S_TYPE(a,b) ((a)&(b))
#define NNGFS_I(ptr) container_of(ptr,struct nngfs_inode_info,vino)

typedef unsigned short unsigned16;
typedef unsigned int unsigned32;

struct nngfs_inode_info  {
        unsigned32 inode;
        unsigned32 blockno; // Block number where data resides. No file is more than a block size.= inode+1
        unsigned16 filetype;
        unsigned16 uid;
        unsigned16 gid;
        unsigned16 size;
        unsigned16 perm;
        struct timespec  atime;
        struct timespec  ctime;
        struct timespec  mtime;
        unsigned32 magic;
        struct inode vino;
};

struct nngfs_inode {
        unsigned32 inode;
        unsigned32 blockno; // Block number where data resides. No file is more than a block size.
        unsigned16 filetype;
        unsigned16 uid;
        unsigned16 gid;
        unsigned16 size;
        unsigned16 perm;
        struct timespec  atime;
        struct timespec  ctime;
        struct timespec  mtime;
        unsigned32 magic;
};

struct nngfs_superblock {
        unsigned32 magic;
        unsigned32 block_bitmap;
        unsigned32 inode_bitmap; //Restricts max number of files to 32
        unsigned32 dirty;
        unsigned16 blocksize;
        unsigned32 root_inode;
	unsigned32 inodes; //File system node counts. Should be 32 for our implementation
	unsigned32 blocks;  //Filesystem size in blocks;
        unsigned16 free_blocks;  // We dont need a free inode counter as 1 inode can have only 1 block max 
        unsigned16 first_data_block;
};
	
/* This is the in memory super block */
struct nngfs_info {
        unsigned32 mount_opt;
	struct nngfs_superblock *ngfs_sb;
	struct buffer_head *sbh; /* Buffer containing the superblock */
};

struct nngfs_dir_entry {
   	unsigned32 inode;
	unsigned32 rec_len;
	unsigned32 name_len;
	char name[NNGFS_NAME_LEN];
};

typedef struct {
 unsigned32 *p;
 unsigned32 key; 
 struct buffer_head *bh; 
} Indirect;

static inline void add_chain(Indirect *p, struct buffer_head *bh, __le32 *v)
{
 p->key = *(p->p = v);
 p->bh = bh;
}

#define setbit(a,b) ((a)|(1<<(b-1)))
#define clearbit(a,b)((a)&(~(1<<(b-1)))) // The bit must be set before

#endif

