/* 
 * nngfs/util/mkngfs.c
 * 
 * Copyright (C) 2008
 * Manish Katiyar (mkatiyar@gmail.com) 
 *
 * NNGFS filesystem creation utility distributed under GPL
 */

#include<stdio.h>
#include<string.h>
#include<stdlib.h>
#include<errno.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<unistd.h>
#include<fcntl.h>
#include<time.h>
#include "nngfs.h"
#define MAX_NAME_LEN 100
extern int errno;
char progname[MAX_NAME_LEN];

static inline unsigned32 setbit(unsigned32 bitmap,int bit){
	return (bitmap|(1<<(bit-1)));
}

void nngfs_create_rootdir_entries(struct nngfs_inode root,int blocksize,int fd){
	struct nngfs_dir_entry rd;
	int ret,pos;
        int start=blocksize*root.blockno;

        // Write the first root entry "."
	memset(&rd,0,sizeof(rd));
	rd.inode = root.inode;
	rd.name_len = strlen(".");
	strcpy(rd.name,".");
	rd.rec_len = sizeof(struct nngfs_dir_entry);
	pos=lseek(fd,start,SEEK_SET);
	if((ret=write(fd,&rd,sizeof(rd)))!=sizeof(rd)){
	 perror("Could not write root dentry entries :");
	 return;
	}

        // Write the second root entry ".."
	pos=lseek(fd,start+rd.rec_len,SEEK_SET);
	memset(&rd,0,sizeof(rd));
	rd.inode = root.inode;
	rd.name_len = strlen("..");
	strcpy(rd.name,"..");
	rd.rec_len = sizeof(struct nngfs_dir_entry);
	if((ret=write(fd,&rd,sizeof(rd)))!=sizeof(rd)){
	 perror("Could not write root dentry entries :");
	 return;
	}

	return;
}

int create_root(int fd,int blocksize){
	int ret;
	struct nngfs_inode root;
        time_t tm;
	memset(&root,0,sizeof(root));
	root.inode = NNGFS_ROOT_INODE; /* Equal to 1 */
	root.blockno = NNGFS_ROOT_INODE+1; /* 0th block is for superblock, First block is for inode info */
	root.filetype = NNGFS_DIR;
	root.uid = root.gid = 0;
	root.size = blocksize;
	root.perm = 0764;
        time(&tm);
        root.atime.tv_sec = root.mtime.tv_sec = root.ctime.tv_sec = tm;
        root.magic = 0xdeadbeef;
	lseek(fd,blocksize*NNGFS_INODE_BLOCK,SEEK_SET);
	if((ret=write(fd,&root,sizeof(root)))!=sizeof(root)){
	 perror("Could not create root inode :");
	 return -1;
	}
	nngfs_create_rootdir_entries(root,blocksize,fd);
	return 0;
}

void create_NNGFS(char *device,int bs){
	struct nngfs_superblock nsb;
	int fd=-1,ret;
	if((fd=open(device,O_WRONLY))==-1){
	 perror("Could not open device :");
	 exit(fd);
	}
	memset(&nsb,0,sizeof(nsb));
	nsb.magic = NNGFS_MAGIC;
	nsb.dirty = NNGFS_CLEAN;
	nsb.inode_bitmap=setbit(nsb.inode_bitmap,NNGFS_ROOT_INODE); //Inode bit set for Root inode = 1
	nsb.block_bitmap=setbit(nsb.block_bitmap,NNGFS_ROOT_BLOCK); //Block bit set for Root inode = 1. Block and inodes have 1:1 mapping
	bs = 0 ? 4096 : bs;
	nsb.blocksize = bs;
	nsb.root_inode = NNGFS_ROOT_INODE;
	nsb.inodes = nsb.blocks = 8*(sizeof(unsigned32));
	nsb.free_blocks = 8*(sizeof(unsigned32)) -1;
	nsb.first_data_block = NNGFS_ROOT_BLOCK;
	if((ret=write(fd,&nsb,sizeof(nsb)))!=sizeof(nsb)){
	 perror("Could not create filesystem :");
         close(fd);
	 exit(fd);
	}
	else{
	 if(create_root(fd,nsb.blocksize)==-1){
	  perror("Could not create filesystem :");
          close(fd);
	  exit(fd);
	 }
         fprintf(stderr,"NNGFS created successfully on \"%s\" of %d blocksize\n",device,bs);
        }
        close(fd);
	return;
}

int main(int argc,char *argv[]){
	char choice,device[MAX_NAME_LEN];
	int blocksize=argv[1] ? atoi(argv[1]) : 4096;
	memset(device,0,MAX_NAME_LEN);
	strcpy(progname,argv[0]);
	printf("\nThis utility creates a NNGFS filesystem type \n\n");
	printf("\tEnter the path of device on which to create NNGFS :");
	scanf("%[^\n]s",device);
	if(!strcmp(device,"")){
         fprintf(stderr,"No device name specified ...!\n");
	 exit(-1);
        }
	printf("\tCreate NNGFS on \"%s\" (Y/N) [default Y] :",device);
	scanf("%c%c",&choice,&choice);
	switch(choice){
	  case 'y':
	  case '\n':
	  case 'Y':
		   create_NNGFS(device,blocksize);
		   break;
          default : printf("Exiting ... !!\n");
	};
	return 0;
}

